/*
 * The Ultimate Logging Module (well, probably not any more)
 *
 * Provides speed enhancing features such as lazy date/time creation,
 * Cacheing of log entries for fast disk access.
 * Lines are NOT written out incomplete, so that multiple agents may
 * log to the same log file without getting in a complete mess.
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:16:02 $
 * ---------------------------------------------------------------------------
 */

#include <phlib/xl_config.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_malloc.h>
#include <phlib/xl_date.h>
#include <phlib/xl_logging.h>

/* Define signal trapped to force cache flushing..*/
#define SIG_FLUSH  SIGHUP

/* Define some constant strings */
#define LINEFEED "\n"
#define SPACE    " "

static CString spaceCS = CSInit;
static CString lfCS    = CSInit;
static CString timeCS  = CSInit;
static CString dateCS  = CSInit;
static CString pidCS   = CSInit;

#define HANDLE_MAX 50

/* Array to hold handles.  Max HANDLE_MAX, probably more than enough */
static TraceHandle handle_list[ HANDLE_MAX ];
static int h_list_ix = 0;

/* Prototypes */
static RetValue allocate_mem( long csize, struct loginfo **logblk );
static RETSIGTYPE sig_flush_cache( int sig );
 
/* ----------------------------------------------------------------------- */

static RetValue
allocate_mem( long csize, struct loginfo **logblk )
{
  uchar_t*    mymem;
  RetValue ret = Failure;

  if( (mymem = (uchar_t*)xl_malloc( sizeof(struct loginfo) )) != NULL ) {

    *logblk = (struct loginfo *)mymem;

    /* Create a logcache of the desired size
	 */
    CreateCString( &((*logblk)->Cache), csize );

    ret = Success;
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
into_cache( struct loginfo *logblk, char *data )
{
  CString dataCS = CSInit;
  CString tmpCS = CSInit;

  xl_get_time( 1 ); /* Read the current time.  Lazy function. 1 = re-read */

  WrapCString( tmpCS, "" );            /* Initialise CString */

  append_cstring( &tmpCS, &dateCS );   /* Append current date    */
  append_cstring( &tmpCS, &spaceCS );  /* Append space character */

  append_cstring( &tmpCS, &timeCS );   /* Append current time    */
  append_cstring( &tmpCS, &spaceCS );  /* Append space character */

  append_cstring( &tmpCS, &pidCS   );  /* Append PID             */
  append_cstring( &tmpCS, &spaceCS );  /* Append space character */
  
  WrapCString( dataCS, data );         /* Wrap up passed in data */
  append_cstring( &tmpCS, &dataCS );   /* Append data to cache   */
  append_cstring( &tmpCS, &lfCS );     /* Append linefeed char   */

  /*
   * There is a special mode of logging where everything is written to an
   * ever increasing buffer in the absence of a file to output to.
   * When would this ever happen?  When you're booting a system that wants
   * to generate log output before it has found out what the logfile is
   * going to be called!
   * Here we test for that mode and write our logline to the file (via a small
   * cache to reduce IO load) or to a unbounded memory buffer.
   */

  if( logblk->fd < 0 ) {
    append_cstring( &logblk->rambuffer, &tmpCS );
  } else {
    /* If we don't have enough room for line in cache..... */
    if( CSLen(tmpCS) >= ( CSMax(logblk->Cache) - CSLen(logblk->Cache) ) ) {
  
      /* Write out cache */
      if( CSLen(logblk->Cache) > 0 ) {
        write( logblk->fd, CSStart(logblk->Cache), CSLen(logblk->Cache) );
      }
  
      /* Clear cache */
      reset_cstring( &(logblk->Cache) );
    }

    if( CSLen(tmpCS) < CSMax(logblk->Cache) ) {
      /* Append new line to Cache */
      append_cstring( &(logblk->Cache), &tmpCS );
    } else {
      /* Cache not big enough for line, so just write it out.. */
      write( logblk->fd, CSStart(tmpCS), CSLen(tmpCS) );
    }
  }

  free_cstring( &tmpCS );

  return Success;
}


/* ----------------------------------------------------------------------- */

TraceHandle
init_ulflog( int trlev, long csize, char *logfile )
{
  struct loginfo *lblk = NULL;
  TraceHandle    lhand;
  int            fd    = -1;

  lhand = 0;

  if( h_list_ix > HANDLE_MAX ) {
    printf( "Too many log files open.  Limit %d\n", HANDLE_MAX );
    return lhand;
  }

  /* if we've been given a logfilename, then open it
   */
  if( logfile != NULL ) {
    if( (fd = open( logfile, O_WRONLY | O_CREAT | O_APPEND | O_NONBLOCK, 0600 )) <= 0 ) {
	  /* Could not open logfile */
	  return (TraceHandle)NULL;
    } 
  }

  if( allocate_mem( csize, &lblk ) == Success ) {

    lblk->trlev     = trlev;
    lblk->fd        = fd;
	WrapCStringLen( lblk->rambuffer, "", 0 );

    /* 
     * Generate handle - Hey, just return alloced mem ptr!
     * Why not? no one will spot it, and it's unique too! 
     * And.... we don't have to bother stacking the handle/ptr
     * pairs because they are the same... just passing the handle
     * gives us the ptr to the memory block and so everything is
     * suddenly all hunky doory!  * s m i l e *
     */
    lhand = (TraceHandle)lblk;

    /* 
     * Store handle in list so we always know where everything is!
     * h_list_ix always points to the next available slot.
     */
    handle_list[ h_list_ix++ ] = lhand;

    /* Initialise date/time functions */
    xl_init_date();
    
    /* Wrap CString around time.  Time is a constant string so
     * doing this is legal.  We've only just read the date/time so
     * we don't bother to do it again here, so pass the 0 flag!
     */
    WrapCString( dateCS, (uchar_t *)xl_get_date(0) );
    WrapCString( timeCS, (uchar_t *)xl_get_time(0) );
  }

  WrapCStringLen( spaceCS, SPACE, 1 ); /* Wrap up a SPACE char    */
  WrapCStringLen( lfCS, LINEFEED, 1 ); /* Wrap up a LINEFEED char */
  IntToCS( &pidCS, getpid() );

  return lhand;
}

/* ----------------------------------------------------------------------- */

void
set_ulflog_level( TraceHandle th, int level )
{
  struct loginfo *lblk = (struct loginfo *)th;
  lblk->trlev = level;
}

/* ----------------------------------------------------------------------- */

void
set_ulflog_cache_size( TraceHandle th, int cachesize )
{
  struct loginfo *lblk = (struct loginfo *)th;

  flush_ulflog();

  free_cstring( &(lblk->Cache) );

  CreateCString( &(lblk->Cache), cachesize );
}

/* ----------------------------------------------------------------------- */

int
get_ulflog_fd( TraceHandle th )
{
  struct loginfo *lblk = (struct loginfo *)th;
  return lblk->fd;
}

/* ----------------------------------------------------------------------- */

RetValue
reopen_ulflog( TraceHandle th, char *logfile  )
{
  struct loginfo *lblk = (struct loginfo *)th;
  int            fd    = -1;

  if( (fd = open( logfile, O_WRONLY | O_CREAT | O_APPEND | O_NONBLOCK, 0600 )) <= 0 ) {
    /* Could not open logfile */
    return Failure;
  }

  lblk->fd = fd;

  /* If we've got any info in the rambuffer, write this straight out to the
   * newly opened log and free the rambuffer.
   */
  if( CSLen(lblk->rambuffer) ) {
    write( lblk->fd, CSStart(lblk->rambuffer), CSLen(lblk->rambuffer) );
	free_cstring( &lblk->rambuffer );
  }

  return Success;
}

/* ----------------------------------------------------------------------- */

RetValue
finish_ulflog( TraceHandle lhand )
{
  struct loginfo *lblk;
  int      i;
  RetValue ret = Failure;

  if( lhand > 0 && h_list_ix > 0 ) {

    /* Clear handle from our list */
    if( h_list_ix == 1 ) {
      handle_list[ --h_list_ix ] = 0;

    } else {

      for(i = 0; i < h_list_ix; i++ ) {
	
	if( lhand == handle_list[ i ] ) {
	  handle_list[i] = handle_list[ --h_list_ix ];
	  break;
	}
      }
    } 

    lblk = (struct loginfo*)lhand; /* Ooh so nice */

    /* Right, before we do anything else we MUST  * F L U S H *  the cache! */
    if( (lblk->Cache).len > 0 ) {
      write( lblk->fd, CSStart(lblk->Cache), CSLen(lblk->Cache) );
    }
   
    /* Close logfile if open */
    if( lblk->fd >= 0 ) close( lblk->fd );
    
	free_cstring( &(lblk->rambuffer) );
	free_cstring( &(lblk->Cache) );

    /* Drop Memory.... */
    xl_free( (char*)lblk );

    ret = Success;
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

void        
ulflog( TraceHandle lhand, int trlev, char *text )
{
  struct loginfo *lblk;

  if( lhand > 0  ) {
    lblk = (struct loginfo *)lhand;
    
    if( trlev <= lblk->trlev ) {  /* Sufficient log level */

      into_cache( lblk, text );
    }
  }
}

/* ----------------------------------------------------------------------- */

#ifdef HAVE_VPRINTF 

void
ulflogf( TraceHandle lhand, int trlev, char *format, ... )
{
  struct loginfo *lblk;
  static char tbuf[ 16384 ];
  va_list ap;
  
  if( lhand > 0  ) {
    lblk = (struct loginfo *)lhand;
    
    if( trlev <= lblk->trlev ) {  /* Sufficient log level */

      /* 
       * I don't like this because we may be creating an output string
       * which is greater than 16384 bytes.... if we do then POW!
       * we break our buffer and it'll all go eugh!
       */
      va_start(ap,format);
      vsprintf( tbuf, format, ap );
      va_end(ap);

      into_cache( lblk, tbuf );
    }
  }
}

#endif

/* ----------------------------------------------------------------------- */

void
flush_ulflog()
{
  int i;
  struct loginfo *lblk;

  for( i = 0; i < h_list_ix; i++ ) {

    /* Take each log cache in turn.... */
    lblk = (struct loginfo *)( handle_list[ i ] );
    
    /* Write out cache */
    write( lblk->fd, (lblk->Cache).s, (lblk->Cache).len );

    /* Clear cache */
    reset_cstring( &(lblk->Cache) );
  }
}

/* ----------------------------------------------------------------------- */

static RETSIGTYPE
sig_flush_cache( int sig )
{
  flush_ulflog();
}

/* ----------------------------------------------------------------------- */
/* end */
