/*
 * ------------------------------------------------------------------------
 * Module : String handling functions
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.3 $  $Date: 2003/10/20 18:54:57 $
 *
 * $Id: xl_string.c,v 1.3 2003/10/20 18:54:57 csmith Exp $
 * $Revision: 1.3 $  $Date: 2003/10/20 18:54:57 $
 *
 * ----------------------------------------------------------------------- 
 */

#include <phlib/xl_config.h>
#include <stdlib.h>
#include <sys/types.h>

#if STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_malloc.h>

/* ----------------------------------------------------------------------- */

int
l_strncmp( const char* s1, const char *s2, int lmax )
{
  int res = 1;
  int i; 
  int len1, len2;
  
  len1 = strlen((char*)s1);
  len2 = strlen((char*)s2); 

  len2 = (len2>lmax) ? lmax : len2;
  len1 = (len1>lmax) ? lmax : len1;
  
  if( len1 == len2 ) {
    for( i = 0; i < len1; i++ ) {
      if( s1[i] != s2[i] ) {
	if( s2[i] > s1[i] ) res = (-1);
	break;
      }
    }
    if( i >= len1 ) res = 0;
  } else if( len2 > len1 ) res = (-1);
  
  return res;
}

/* ----------------------------------------------------------------------- */
/*
 * Takes string of type CString and blits it uppercase
 * Entry: CString - ** Modified **!
 * Exit : CString in Upper Case.
 */

void
upper_case( CString *str )
{
  int i;
  for( i = 0; i < str->len; i++ ) str->s[i] = ToUpper( str->s[i] );
}

/* ----------------------------------------------------------------------- */
/*
 * Takes a CString and chops off the first word, which it returns.
 *
 * Entry: Pointer to destingation CString - ** Modified, data not! **
 *        Pointer to source CString       - ** Modified **
 * Exit : Returns a CString containing the chopped word
 *        Returns Success if a chop was made, failure otherwise.
 */

RetValue
chop_word( CString *to, CString *str )
{
  int i;
  RetValue ret = Failure;
  char *sp, *ep;
  ep = str->s;
  
  free_cstring( to );
  if(  str->len  < 1 ) return Failure;

  /* Search for end of current word */
  sp = ep;
  for(i = 0 ; i < str->len &&  *ep != ' '; ep++, i++ );
  
  /* Set str to point to beginning of next word 
   * (Which _may_ include leading space)
   */
  if( i >= str->len ) { /* There was only one word in the string! */
    str->len = 0; 
    str->s   = ep; 
  } else {
    str->len -= (i + 1);
    str->s    = ep + 1;  /* +1 to move passed sep char */
  }
  ret = Success;
  
  /* Create new CString */
  WrapCStringPtrLen( to, sp, ep - sp );

  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * Takes a CString and chops off the first string up to the separater
 * character, which it returns (minus the separator).
 *
 * Entry: Pointer to destination CString - ** Modified, data not! **
 *        Pointer to source CString      - ** Modified **
 * Exit : Returns a CString containing the chopped line (string)
 *        Returns Success if a chop was made, failure otherwise.
 */

RetValue
chop_line( CString *to, CString *str, char sep )
{
  int i;
  char *sp, *ep;
  ep = str->s;
  
  free_cstring( to );
  if(  str->len  < 1 ) return Failure;
  
  /* Search for separator character */
  sp = ep;
  for(i = 0 ; i < str->len &&  *ep != sep; ep++, i++ );
  
  /* Set str to point to beginning of next word 
   * (Which _may_ include leading space)
   */
  if( i >= str->len ) { /* There was only one 'line' in the string! */
    str->len = 0; 
    str->s   = ep;
  } else {
    str->len -= (i + 1);
    str->s    = ep + 1;  /* +1 to move passed sep char */
  }
  /* Create new CString */
  WrapCStringPtrLen( to, sp, ep - sp );

  return Success;

}

/* ----------------------------------------------------------------------- */
/*
 * Appends a CString onto the end of another CString
 *
 * Entry: Pointer to Destination CString, pointer to CString to append
 * Exit : Dest CString will be free'd if it isn't a ConstString.
 *          Space for the new string is malloced and the Dest CString is
 *          set to point to this area.
 *
 * Returns Success or Failure
 *
 * Notes: Any string which is appended to should be passed into free_cstring()
 *          once it is no longer needed.
 */
RetValue 
append_cstring( CString *to, const CString *from )
{
  char *buf;
  uint32 req;
  uint32 size, ol;

  if ( from->len < 1 ) return Success; /* No error, nothing  to append */

  if( (req = to->len + from->len) >= to->max || to->f == 'n') {
    /* We need to grow space for string  -- Move to the next power of two */
    size = 32;
    while( size <= req ) size<<=1;
    buf = xl_malloc( size + 1 );

    /* This was from orchid code - It's a bit pooey becasue we should _always_
     * try and reserve 2^n+1 where 2^n <= required  bytes, AND NOT
     *                 2^n+1 where 2^n <= previously reserved bytes.
     * the latter case means that we may _still_ not have enough memory and
     * need to do a req * 2 to get enough + some more.
     *	int z = 32;
     *  while( z <= to->max) z<<=1;
     *  size = z;
     *  if( size < req ) size = req * 2;
     *  buf = xl_malloc( size + 1 );
     */

#ifdef DEVELOPMENT
    printf( "Require %ld... Mallocing %ld\n", (uint32)req, (uint32)size );
    printf( "append_cstring: Malloced Ptr: %x\n", (uint32)buf );
#endif
    
    if( buf == NULL ) return Failure; /* Error condition */
    
    /* copy string */
    ol = to->len;
    memcpy( buf, to->s, ol );
    if( to->f == 'y' ) {
#ifdef DEVELOPMENT
      printf( "append_cstring: Not big enough.. "
	      "Freeing old space: %x\n", (uint32)to->p );
#endif
      free_cstring( to );
    }
    to->p = to->s = buf;
    to->max = size;
    to->len = ol;
    to->f   = 'y';     /* Say that space was malloced */

  } else { /* New string fits, but do we need to shift first string left? */
#ifdef DEVELOPMENT
    printf( "append_cstring: Check slide\n" );
#endif
    if( (to->s + to->len + from->len) > (to->p + to->max) ) {
#ifdef DEVELOPMENT
      printf( "append_cstring:   Sliding original string\n" );
#endif
      memcpy( to->p, to->s, to->len );
      to->s = to->p;
    }
  }  
  
  /* Now append string - if src data ptr is NULL, then append nothing and
   * do not extend the length of the existing string.  In effect we've just
   * added space to the end of the original string.  This only happens if we
   * call extend_cstring().
   */
  if( from->s ) {
    memcpy( to->s + to->len, from->s, from->len );
    to->len += from->len;
  }	
  
  return Success;
}

/* ----------------------------------------------------------------------- */
/* This function takes a CString, mallocs an area big enough to hold the
 * data currently pointed to and copies the data into the new area. It is 
 * then set to point to this new data.  The previous data area is 'forgotten'
 * about - So be very carefull to keep a pointer to the previouse data area
 * so that you can free it up later etc.....
 *
 * Entry: CString to become a clone
 * Exit : CString which looks exactly like the original, but is a clone
 *        Returns Success or Failure
 */ 
RetValue
CloneCString( CString *str )
{
  RetValue res = Failure;
  char *new = xl_malloc( str->len  + 1);
  
#ifdef DEVELOPMENT
  printf( "CloneCString: Malloc ptr: %x\n", (uint32)new );
#endif

  if( new != NULL ) {
    res = Success;
    memcpy( new, str->s, str->len );
    str->p = str->s = new;
    str->max = str->len;
    str->f = 'y';
  }
  return res;
}

/* ----------------------------------------------------------------------- */

void 
free_cstring( CString *str )
{
  if( str->f == 'y' && str->p != NULL ) {
#ifdef DEVELOPMENT
    printf( "free_cstring: Free ptr: %x             StringPtr: %x\n", 
	    (uint32)str->p,  (uint32)str );
#endif
    xl_free( str->p );
    str->p = NULL;
  }
#ifdef DEVELOPMENT
  else printf( "free_cstring: Not freeing\n" );
#endif
  str->len = str->max = 0;
  str->f = 'n';
  str->p = str->s = 0;
}

/* ----------------------------------------------------------------------- */
/*
 * Extends the internal space of a cstring by n bytes.
 * Does NOT extend the value returned by CSLen() as we're NOT adding data
 *  to the cstring - we're extending CSMax().
 *
 * Entry: Pointer to CString to extend - ** Modified **
 *        Size to reserve.
 * Exit : Memory malloced and the passed CString wrapped around it.
 *        Returns Success or Failure
 *
 * Notes: Any string returned from this function must be passed into 
 *        free_cstring() once it is no longer needed.
 */ 
RetValue
extend_cstring( CString *str, int size )
{
  CString eCS = CSInit;
  
  eCS.p    = eCS.s = NULL;
  eCS.max  = 0;
  eCS.f    = 'n';
  eCS.len  = size;
  
  return append_cstring( str, &eCS );
}

/* ----------------------------------------------------------------------- */
/*
 * Mallocs space for a CString and wraps such a string around it...
 * Entry: Pointer to destination CString - ** Modified **
 *        Size to reserve.
 * Exit : Memory malloced and the passed CString wrapped around it.
 *        Returns Success or Failure
 *
 * Notes: Any string which is 'Created' should be passed into free_cstring()
 *          once it is no longer needed.
 */ 
RetValue
CreateCString( CString *str, int size )
{
  char *buf;
  /* free_cstring( str ); */
  
  if( (buf = xl_malloc( size + 1 ) ) == NULL ) return Failure;
  
#ifdef DEVELOPMENT
  printf( "CreateCString: Malloc ptr: %x\n", (uint32)buf );
#endif

  str->p    = str->s = buf;
  str->max  = size;
  str->f    = 'y';
  str->len  = 0;
  
  return Success;
}

/* ----------------------------------------------------------------------- */

/* 
 * Removes leading spaces from a CString
 *
 * Entry: Pointer to CString
 * Exit : CString updated to skip leading space.  Data is not modified
 */
void
skip_lead_space( CString *str )
{
  int i = 0;
  for( i = 0; i < str->len && *str->s == ' '; str->s++, i++ );
  str->len -= i;
}	

/* ----------------------------------------------------------------------- */
/*
 * Removes trailing spaces from end of CString
 *
 * Entry: Pointer to CString
 * Exit : CString updated to avoid trailing space. - ** Modified **
 *        It considers both ' ' and '\0' characters to be 'spaces'
 */
void
skip_tail_space( CString *str )
{
  int i = 0;
  char *e = str->s + str->len - 1;
  
  for( i = 0; i < str->len && (*e == ' ' || *e == '\0'); e--, i++ );
  str->len -= i;
}	

/* ----------------------------------------------------------------------- */
/*
 * Entry: CString to test,
 *        char string to compare;
 * Exit : Returns 0 if CString does not end with charstring
 *                1 if it does
 */
int
ends_with_chars( const CString *text, const char* str )
{
  char *s1, *s2;
  int res = 1;
  int len = strlen( (char*)str );
  int i;

  if( len == 0 || len > text->len ) return 0; 

  s1 = text->s + text->len - len;
  s2 = (char*)str;

  for( i = 0; i < len; i++ ) {
    if( s1[i] != s2[i] ) {
      res = 0; /* Failure */
      break;
    }
  }
  
  return res;
}

/* ----------------------------------------------------------------------- */
/*
 * Entry: CString to test,
 *        char string to compare;
 * Exit : Returns 0 if CString does not begin with charstring
 *                1 if it does
 */
int
begins_with_chars( const CString *text, const char* str )
{
  char *s1, *s2;
  int res = 1;
  int len = strlen( (char*)str );
  int i;

  if( len == 0 || len > text->len ) return 0; 

  s1 = text->s;
  s2 = (char*)str;

  for( i = 0; i < len; i++ ) {
    if( s1[i] != s2[i] ) {
      res = 0; /* Failure */
      break;
    }
  }
  
  return res;
}

/* ----------------------------------------------------------------------- */
/* Keeps hold of the Left most n characters of string 
 *
 * Entry: ptr to CString to modify,
 *        Number of characters to keep hold of
 * Exit : CString is modified,
 *        Returns 1 if function succeeded
 *                0 if the required size is BIGGER then the original len
 */
int 
string_left( CString *str, int size )
{
  int res = 0;
  
  if( str->len > size && size >= 0 ) {
    str->len = size;
    res = 1;
  }
  return res;
}


/* ----------------------------------------------------------------------- */
/* 
 * Reset start pointer and length of a CString back to the start...
 */
void
reset_cstring( CString *str )
{
  str->len = 0;
  str->s   = str->p;
}

/* ----------------------------------------------------------------------- */
/* Keeps hold of the Right most n characters of string 
 *
 * Entry: ptr to CString to modify,
 *        Number of characters to keep hold of
 * Exit : CString is modified,
 *        Returns 1 if function succeeded
 *                0 if the required size is BIGGER then the original len
 */
int 
string_right( CString *str, int size )
{
  int res = 0;
  
  if( str->len > size && size >= 0 ) {
    str->s += (str->len - size);
    str->len = size;
    res = 1;
  }
  return res;
}


/* ----------------------------------------------------------------------- */
/* This function keeps hold of a subset of a cstring
 *
 * Entry: ptr to CString to affect
 *        Start  character index
 *        Finish character index
 * Exit : CString modified
 *        Returns 1 if function succeeded
 *                0 if either Start or Finish outside bounds of original str
 */
int
string_mid( CString *str, int st, int fin )
{
  int res = 0;
  if( str->len > st && str->len >= fin && 
      st >= 0 && fin >= 0 && fin >= st ) {
    str->s += st;
    str->len = ( fin - st );
    res = 1;
  }
  return res;
}

/* ----------------------------------------------------------------------- */
/* This function extracts a subset of a cstring into _another_ CString
 *
 * Entry: ptr to CString to SET
 *        ptr to CString to read from
 *        Start  character index
 *        Finish character index
 * Exit : CString set to substring of source
 *        Returns 1 if function succeeded
 *                0 if either Start or Finish outside bounds of original str
 */
int
get_string_mid( CString *dst, CString *str, int st, int fin )
{
  int res = 0;
  if( str->len > st && str->len >= fin && 
      st >= 0 && fin >= 0 && fin >= st ) {
    
    free_cstring( dst ); /* Very Very Necessary!! - free's and initialises!! */

    dst->s = dst->p = str->s + st;
    dst->len = dst->max = ( fin - st );
    res = 1;
  }
  return res;
}

/* ----------------------------------------------------------------------- */
/* This function compares CStrings
 * Entry: CStrings to compare
 * Exit : 0 if the srings compare,
 *        1 if s1 > s2
 *       -1 if s1 < s2
 */
int
CScmp( CString *s1, CString *s2 )
{
  int res = 1;
  int i; 
  
  if( s1->len == s2->len ) {
    for( i = 0; i < s1->len; i++ ) {
      if( s1->s[i] != s2->s[i] ) {
	if( s2->s[i] > s1->s[i] ) res = (-1);
	break;
      }
    }
    if( i >= s1->len ) res = 0;
  } else if( s2->len > s1->len ) res = (-1);
  
  return res;
}

/* ----------------------------------------------------------------------- */
/* This function compares CStrings alphabetically
 * Entry: CStrings to compare
 * Exit : 0 if the srings compare,
 *        1 if s1 > s2
 *       -1 if s1 < s2
 */
int
CScmpAlphabetic( CString *s1, CString *s2 )
{
  int res = 0;
  int i; 
  char c1, c2;
  
  for( i = 0; i < s1->len; i++ ) {

    c1 = (i < s1->len) ? s1->s[i] : 0 ;
    c2 = (i < s2->len) ? s2->s[i] : 0 ;

    if( c1 < c2 ) { res = (-1); break; }
    else if( c1 > c2 ) { res = 1; break; }
  }
  
  return res;
}

/* ----------------------------------------------------------------------- */
/* Function copies data from a CString into a character array of max size
 * max_size.  The character array is then NULL terminated.  If the Cstring
 * is too big for the character array, then the leading characters are
 * dropped. 
 */   
void
CStoCharCpy( char *chs, CString *cstr, int max_size )
{
  int os = (cstr->len < max_size ) ? 0 : cstr->len - max_size + 1;
  
  memcpy( chs, cstr->s + os, cstr->len - os );
  chs[ cstr->len - os ] = '\0';
}

/* ----------------------------------------------------------------------- */

RetValue
IntToCS( CString* dst, int val )
{
  char* buf;
  int   a   = val;
  int   neg = 0;
  int   l   = 0;

  if( a < 0 ) { /* If a negative number, adjust value and set flag */
    a *= (-1);
    neg = 1;
  }

  if( l == 0 && a >= 100000000 ) l += 9;
  if( l == 0 && a >= 10000000  ) l += 8;
  if( l == 0 && a >= 1000000   ) l += 7;
  if( l == 0 && a >= 100000    ) l += 6;
  if( l == 0 && a >= 10000     ) l += 5;
  if( l == 0 && a >= 1000      ) l += 4;
  if( l == 0 && a >= 100       ) l += 3;
  if( l == 0 && a >= 10        ) l += 2;
  if( l == 0 && a >= 0         ) l += 1;

  l += neg; /* Add neg flag (which is 1 if we need room of -ve sign) */
  if( dst->max < l ) {    /* May as well allocate a chunk */

    buf = xl_malloc( 128 );
#ifdef DEVELOPMENT
    printf( "IntToCS: Malloc ptr: %x\n", (uint32)buf );
#endif
    if( buf == NULL ) return Failure; /* Error condition */
    if( dst->f == 'y' ) free_cstring( dst );
    
    dst->p = dst->s = buf;
    dst->max = 128;
    dst->f   = 'y';     /* Say that space was malloced */
  }

  dst->len = l;

  l--; /* index from 0 */
  do {
    dst->s[l] = a % 10 + '0';
    a /= 10;
    l--;
  } while( a > 0 );
  if( neg == 1 ) dst->s[l] = '-';

  return Success;
}

/* ----------------------------------------------------------------------- */

RetValue
CStoInt32( CString *src, int32 *val )
{
  RetValue ret = Failure;

  register int c, i = 0;
  register int32 v = 0;
  register char *p = CSStart( (*src) );
  int l = CSLen( (*src) );
  int m = 1;

  if( *p == '-' ) { p++; l--; m = MINUSONE; }

  if(l>9) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*1000000000); else l=-1; }
  if(l>8) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*100000000 ); else l=-1; }
  if(l>7) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*10000000  ); else l=-1; }
  if(l>6) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*1000000   ); else l=-1; }
  if(l>5) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*100000    ); else l=-1; }
  if(l>4) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*10000     ); else l=-1; }
  if(l>3) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*1000      ); else l=-1; }
  if(l>2) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*100       ); else l=-1; }
  if(l>1) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c*10        ); else l=-1; }
  if(l>0) { c=*(p+i++)-'0'; if(c>=0 && c<=9) v+=(c           ); else l=-1; } 

  if( l > MINUSONE ) {
    ret = Success;
    *val = v * m;
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
append_intCS( CString* dst, int val )
{
  RetValue ret;

  static char buf[12];
  CString tmp;
  WrapCStringLen( tmp, buf, 12 );
  if( (ret = IntToCS( &tmp, val )) == Success )
    ret = append_cstring( dst, &tmp );
  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
is_cstring_blank( CString *str )
{
  RetValue ret = Success;
  int i;

  for(i = 0; i < str->len; i++ ) {
    if( str->s[i] != ' ' ) {
      ret = Failure;
      i = str->len;
    }
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
append_chars( CString *cs, const char *s )
{
  static CString tmpCS;
  
  WrapCString( tmpCS, (char*)s );
  return append_cstring( cs, &tmpCS );
}


/* ----------------------------------------------------------------------- */
/*
 * This function returns the index in src of the first occurrance of
 * string fnd.  If fnd is not a substring of src then MINUSONE is returned.
 * -- This is NOT the fastest piece of code in the world.
 * -- We move through the src string a byte at a time, which is slow.
 * --
 * -- Use boyermoore_cstring( ..., ... ) instead
 */

int
index_cstring( CString *src, CString *fnd )
{
  int ret = MINUSONE;

  char *p1 = src->s;
  char *p2 = fnd->s;
  uint  l1  = src->len;
  uint  l2  = fnd->len;
  uint  ix  = 0;

  while( l1 >= l2) {
    ix = 0;
    while( *(p1+ix) == *(p2+ix) && ix < fnd->len ) ix++;
    if( ix == l2 ) { ret = p1 - src->s; break; }
    p1++;
    l1--;
  }
  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * Takes a CString and chops off the last word, which it returns.
 *
 * Entry: Pointer to destingation CString - ** Modified, data not! **
 *        Pointer to source CString       - ** Modified **
 * Exit : Returns a CString containing the chopped word
 *        Returns Success if a chop was made, failure otherwise.
 */

RetValue
chop_word_last( CString *to, CString *str )
{
  int i;
  RetValue ret = Failure;
  char *sp, *ep;
  ep = str->s + str->len - 1; /* Sit on last byte os cstring */
  
  free_cstring( to );
  if(  str->len  < 1 ) return Failure;

  /* Search back for start of last word */
  sp = ep;
  for(i = 0 ; i < str->len &&  *ep != ' '; ep--, i++ );
  
  /* Loose last word from string
   * (new string _may_ include trailing space)
   */
  if( i >= str->len ) { /* There was only one word in the string! */
    str->len = 0; 
  } else {
    str->len -= (i + 1);
  }
  ret = Success;
  
  /* Create new CString */
  WrapCStringPtrLen( to, ep + 1, sp - ep);

  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * Takes a CString and chops off the LAST string beginning with the separater
 * character, which it returns (minus the separator).
 *
 * Entry: Pointer to destination CString - ** Modified, data not! **
 *        Pointer to source CString      - ** Modified **
 * Exit : Returns a CString containing the chopped line (string)
 *        Returns Success if a chop was made, failure otherwise.
 */

RetValue
chop_line_last( CString *to, CString *str, char sep )
{
  int i;
  RetValue ret = Failure;
  char *sp, *ep;
  ep = str->s + str->len - 1; /* Sit on last byte os cstring */
  
  free_cstring( to );
  if(  str->len  < 1 ) return Failure;

  /* Search back for start of last word */
  sp = ep;
  for(i = 0 ; i < str->len &&  *ep != sep; ep--, i++ ); 
  
  /* Loose last word from string
   * (new string _may_ include trailing space)
   */
  if( i >= str->len ) { /* There was only one word in the string! */
    str->len = 0; 
  } else {
    str->len -= (i + 1);
  }
  ret = Success;
  
  /* Create new CString */
  WrapCStringPtrLen( to, ep + 1, sp - ep);

  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * Do case insensitive compare of a CString and char string 
 */
int
strcmpCS( CString *str1, const char *str2)
{
  int res, i, len1;
  char c1, c2;

  res = 0;
  len1 = CSLenPtr( str1 );
  for(i = 0; i < len1 && res == 0; i++) {
    c1 = CSCharPtr(str1, i);
    if( (c2 = str2[i]) == 0 ) return 1; /* str2 shorter than str1 */
		
    if(c1 >= 'a' && c1 <= 'z') c1 -= 32;
    if(c2 >= 'a' && c2 <= 'z') c2 -= 32;

    if(c1 < c2) res = MINUSONE;
    else if(c1 > c2) res = 1;
  }
  if( res == 0 && str2[i] != 0 ) res = MINUSONE; /* str2 is longer than str1 */

  return res;
}

/* ----------------------------------------------------------------------- */
/*
 * Boyer-Moore string search algorithm
 *

 * Entry: fnd   = String to initialise search table for
 *        mbtab = array of 256 unsigned integers that boyer-moore use as a
 *                search table for the given string
 */

uint *
boyermoore_init( CString *fnd, uint *bmtab )
{
  uint slx = fnd->len - 1;
  int i;
  /* Initialise delta table */
  for( i = 0; i < 256; i++ ) bmtab[i] = fnd->len;
  for( i = 0; i < slx; i++ ) bmtab[ (int)(fnd->s[i]) ] = slx-i;

  return bmtab;
}

/* ----------------------------------------------------------------------- */
/* This function returns the index in src of the first occurrance of
 * string fnd.  If fnd is not a substring of src then MINUSONE is returned.
 *
 * Entry: Src   - document to search
 *        fnd   - string to search for
 *        bmtab - A boyer-moore search table, pre-initialised with the
 *                search string.
 */
int
boyermoore_cstring( CString *src, CString *fnd, uint *bmtab )
{
  const uint flen = fnd->len;
  const uint slen = src->len;
  const uchar_t *f = (uchar_t*)fnd->s;
  const uchar_t *s = (uchar_t*)src->s;
  int six, fix, delta;
  if( ! (flen && slen) ) return MINUSONE;
  /* loop for all chars in fnd */
  for( six = flen-1, fix = flen-1; fix >= 0; six--, fix-- ) {
    while( s[six] != f[fix] ) {
      delta = bmtab[ s[six] ];
      /* Move srch index up by as much as we can.. */
      six += ( flen - fix <= delta ) ? delta : flen - fix;
      if( six >= slen ) return MINUSONE;
      fix = flen - 1;
    }
  }
  return six + 1;
}

/* ----------------------------------------------------------------------- */
/* end */
